"""cenclave.core.test_docker module."""

from pathlib import Path
from typing import ClassVar, Dict, List, Optional, Tuple

from pydantic import BaseModel


class TestDockerConfig(BaseModel):
    """Definition of a running test docker configuration."""

    port: int
    code: Path
    application: str
    sealed_secrets: Optional[Path]
    secrets: Optional[Path]
    simu_enclave_sk: Optional[Path]

    secret_mountpoint: ClassVar[str] = "/root/.cache/cenclave/secrets.json"
    sealed_secrets_mountpoint: ClassVar[str] = (
        "/root/.cache/cenclave/sealed_secrets.json"
    )
    enclave_sk_mountpoint: ClassVar[str] = "/key/enclave.key"

    code_mountpoint: ClassVar[str] = "/app"
    entrypoint: ClassVar[str] = "cenclave-test"

    def cmd(self) -> List[str]:
        """Serialize the docker command args."""
        return ["--application", self.application, "--debug"]

    def ports(self) -> Dict[str, Tuple[str, str]]:
        """Define the docker ports."""
        return {f"{self.port}/tcp": ("127.0.0.1", str(self.port))}

    def volumes(self) -> Dict[str, Dict[str, str]]:
        """Define the docker volumes."""
        v = {
            f"{self.code.resolve()}": {
                "bind": TestDockerConfig.code_mountpoint,
                "mode": "rw",
            },
        }

        if self.secrets:
            v[f"{self.secrets.resolve()}"] = {
                "bind": TestDockerConfig.secret_mountpoint,
                "mode": "rw",
            }

        if self.sealed_secrets:
            v[f"{self.sealed_secrets.resolve()}"] = {
                "bind": TestDockerConfig.sealed_secrets_mountpoint,
                "mode": "rw",
            }

        if self.simu_enclave_sk:
            v[f"{self.simu_enclave_sk.resolve()}"] = {
                "bind": TestDockerConfig.enclave_sk_mountpoint,
                "mode": "rw",
            }

        return v



"""cenclave.core.ignore_file module."""

from pathlib import Path


class IgnoreFile:
    """Class to deal with the .cenclaveignore file."""

    @staticmethod
    def parse(path: Path):
        """Parse the cenclaveignore from `path`."""
        ignore_file = path / ".cenclaveignore"
        return (
            {
                line
                for line in ignore_file.read_text().splitlines()
                # Ignore empty lines
                # Ignore lines being comments
                if line.strip() and not line.strip().startswith("#")
            }
            if ignore_file.exists()
            else {}
        )



"""Test model/docker.py."""


from cenclave.core.sgx_docker import SgxDockerConfig


def test_load():
    """Test `load` function."""
    ref_conf = SgxDockerConfig(
        size=4096,
        host="127.0.0.1",
        port=7788,
        subject="CN=myapp.fr,O=MyApp Company,C=FR,L=Paris,ST=Ile-de-France",
        subject_alternative_name="myapp.fr",
        app_id="4141a3e6-1f2b-4ccf-8610-aa0891a1a210",
        expiration_date=1714639412,
        app_dir="/home/cosmian/workspace/sgx_operator/",
        application="app:app",
        healthcheck="/health",
        signer_key="/opt/cosmian-internal/cosmian-signer-key.pem",
    )

    conf = SgxDockerConfig.load(
        docker_labels={"healthcheck_endpoint": "/health", "cenclave": "1"},
        docker_attrs={
            "HostConfig": {
                "PortBindings": {
                    "443/tcp": [{"HostIp": "127.0.0.1", "HostPort": "7788"}]
                },
            },
            "Config": {
                "Cmd": [
                    "--size",
                    "4096M",
                    "--subject",
                    "CN=myapp.fr,O=MyApp Company,C=FR,L=Paris,ST=Ile-de-France",
                    "--san",
                    "myapp.fr",
                    "--id",
                    "4141a3e6-1f2b-4ccf-8610-aa0891a1a210",
                    "--application",
                    "app:app",
                    "--expiration",
                    "1714639412",
                ],
            },
            "Mounts": [
                {
                    "Type": "bind",
                    "Source": "/opt/cosmian-internal/cosmian-signer-key.pem",
                    "Destination": "/root/.config/gramine/enclave-key.pem",
                    "Mode": "rw",
                    "RW": True,
                    "Propagation": "rprivate",
                },
                {
                    "Type": "bind",
                    "Source": "/home/cosmian/workspace/sgx_operator/",
                    "Destination": "/opt/input",
                    "Mode": "rw",
                    "RW": True,
                    "Propagation": "rprivate",
                },
                {
                    "Type": "bind",
                    "Source": "/var/run/aesmd",
                    "Destination": "/var/run/aesmd",
                    "Mode": "rw",
                    "RW": True,
                    "Propagation": "rprivate",
                },
            ],
        },
    )

    assert conf == ref_conf


def test_labels():
    """Test `labels` function."""
    ref_conf = SgxDockerConfig(
        size=4096,
        host="127.0.0.1",
        port=7788,
        subject="CN=myapp.fr,O=MyApp Company,C=FR,L=Paris,ST=Ile-de-France",
        subject_alternative_name="myapp.fr",
        app_id="4141a3e6-1f2b-4ccf-8610-aa0891a1a210",
        expiration_date=1714639412,
        app_dir="/home/cosmian/workspace/sgx_operator/",
        application="app:app",
        healthcheck="/health",
        signer_key="/opt/cosmian-internal/cosmian-signer-key.pem",
    )

    assert ref_conf.labels() == {"cenclave": "1", "healthcheck_endpoint": "/health"}


def test_devices():
    """Test `devices` function."""
    ref_conf = SgxDockerConfig(
        size=4096,
        host="127.0.0.1",
        port=7788,
        subject="CN=myapp.fr,O=MyApp Company,C=FR,L=Paris,ST=Ile-de-France",
        subject_alternative_name="myapp.fr",
        app_id="4141a3e6-1f2b-4ccf-8610-aa0891a1a210",
        expiration_date=1714639412,
        app_dir="/home/cosmian/workspace/sgx_operator/",
        application="app:app",
        healthcheck="/health",
        signer_key="/opt/cosmian-internal/cosmian-signer-key.pem",
    )

    assert ref_conf.devices() == [
        "/dev/sgx_enclave:/dev/sgx_enclave:rw",
        "/dev/sgx_provision:/dev/sgx_provision:rw",
        "/dev/sgx/enclave:/dev/sgx/enclave:rw",
        "/dev/sgx/provision:/dev/sgx/provision:rw",
    ]


def test_ports():
    """Test `ports` function."""
    ref_conf = SgxDockerConfig(
        size=4096,
        host="127.0.0.1",
        port=7788,
        subject="CN=myapp.fr,O=MyApp Company,C=FR,L=Paris,ST=Ile-de-France",
        subject_alternative_name="myapp.fr",
        app_id="4141a3e6-1f2b-4ccf-8610-aa0891a1a210",
        expiration_date=1714639412,
        app_dir="/home/cosmian/workspace/sgx_operator/code.tar",
        application="app:app",
        healthcheck="/health",
        signer_key="/opt/cosmian-internal/cosmian-signer-key.pem",
    )

    assert ref_conf.ports() == {"443/tcp": ("127.0.0.1", "7788")}


def test_volumes():
    """Test `volumes` function."""
    ref_conf = SgxDockerConfig(
        size=4096,
        host="127.0.0.1",
        port=7788,
        subject="CN=myapp.fr,O=MyApp Company,C=FR,L=Paris,ST=Ile-de-France",
        subject_alternative_name="myapp.fr",
        app_id="4141a3e6-1f2b-4ccf-8610-aa0891a1a210",
        expiration_date=1714639412,
        app_dir="/home/cosmian/workspace/sgx_operator/",
        application="app:app",
        healthcheck="/health",
        signer_key="/opt/cosmian-internal/cosmian-signer-key.pem",
    )

    assert ref_conf.volumes() == {
        "/opt/cosmian-internal/cosmian-signer-key.pem": {
            "bind": "/root/.config/gramine/enclave-key.pem",
            "mode": "rw",
        },
        "/home/cosmian/workspace/sgx_operator": {
            "bind": "/opt/input",
            "mode": "rw",
        },
        "/var/run/aesmd": {
            "bind": "/var/run/aesmd",
            "mode": "rw",
        },
    }


def test_cmd():
    """Test `cmd` function."""
    ref_conf = SgxDockerConfig(
        size=4096,
        host="127.0.0.1",
        port=7788,
        subject="CN=myapp.fr,O=MyApp Company,C=FR,L=Paris,ST=Ile-de-France",
        subject_alternative_name="myapp.fr",
        app_id="4141a3e6-1f2b-4ccf-8610-aa0891a1a210",
        expiration_date=1714639412,
        app_dir="/home/cosmian/workspace/sgx_operator/",
        application="app:app",
        healthcheck="/health",
        signer_key="/opt/cosmian-internal/cosmian-signer-key.pem",
    )

    assert ref_conf.cmd() == [
        "--size",
        "4096M",
        "--subject",
        "CN=myapp.fr,O=MyApp Company,C=FR,L=Paris,ST=Ile-de-France",
        "--san",
        "myapp.fr",
        "--id",
        "4141a3e6-1f2b-4ccf-8610-aa0891a1a210",
        "--application",
        "app:app",
        "--expiration",
        "1714639412",
    ]



"""Test boostrap.py."""

from uuid import UUID

from cenclave.core.bootstrap import ConfigurationPayload


def test_payload():
    """Test `payload` function."""
    conf = ConfigurationPayload(
        app_id=UUID("63322f85-1ff8-4483-91ae-f18d7398d157"),
        secrets=None,
        sealed_secrets=None,
        code_secret_key=None,
    )

    assert conf.payload() == {
        "uuid": "63322f85-1ff8-4483-91ae-f18d7398d157",
    }

    conf = ConfigurationPayload(
        app_id=UUID("63322f85-1ff8-4483-91ae-f18d7398d157"),
        secrets={"key": "password"},
        sealed_secrets=b"123456789",
        code_secret_key=b"abcdefg",
    )

    assert conf.payload() == {
        "uuid": "63322f85-1ff8-4483-91ae-f18d7398d157",
        "app_secrets": {"key": "password"},
        "app_sealed_secrets": "MTIzNDU2Nzg5",
        "code_secret_key": "61626364656667",
    }



"""cenclave.home.command.sgx_operator.spawn module."""

import os
from datetime import datetime, timedelta
from pathlib import Path
from uuid import uuid4

from docker.client import DockerClient
from docker.models.containers import Container

from cenclave.command.helpers import (
    app_container_exists,
    enclave_size_integer,
    get_app_container,
    get_client_docker,
    get_running_app_container,
    is_port_free,
    load_docker_image,
)
from cenclave.command.sgx_operator.evidence import (
    collect_evidence_and_certificate,
    guess_pccs_url,
)
from cenclave.core.bootstrap import wait_for_conf_server
from cenclave.core.clock_tick import ClockTick
from cenclave.core.conf import AppConf, AppConfParsingOption
from cenclave.core.sgx_docker import SgxDockerConfig
from cenclave.core.spinner import Spinner
from cenclave.error import AppContainerAlreadyRunning, AppContainerError, PortBusy
from cenclave.log import LOGGER as LOG
from cenclave.model.package import CodePackage


def add_subparser(subparsers):
    """Define the subcommand."""
    parser = subparsers.add_parser(
        "spawn", help="spawn container with specified web app"
    )

    parser.add_argument(
        "name",
        type=str,
        help="name of the application",
    )

    parser.add_argument(
        "--package",
        type=Path,
        required=True,
        help="tarball containing the container images and the code",
    )

    parser.add_argument(
        "--host",
        type=str,
        default="0.0.0.0",
        help="common name of the generated certificate",
    )

    parser.add_argument(
        "--subject",
        type=str,
        default="CN=cosmian.com,O=Cosmian Tech,C=FR,L=Paris,ST=Ile-de-France",
        help="Subject in the RA-TLS certificate",
    )

    parser.add_argument(
        "--san",
        type=str,
        required=True,
        help="Subject Alternative Name in the RA-TLS certificate",
    )

    parser.add_argument(
        "--days",
        type=int,
        help="number of days before the certificate expires",
        default=365,
    )

    parser.add_argument(
        "--port",
        type=int,
        default=443,
        help="application port",
    )

    parser.add_argument(
        "--size",
        type=enclave_size_integer,
        required=True,
        help="enclave size to spawn (must be a power of 2)",
    )

    parser.add_argument(
        "--client-certificate",
        type=Path,
        help="bundle for client certificate authentication",
    )

    parser.add_argument(
        "--signer-key",
        type=Path,
        help="enclave signer key",
        default=f"{os.getenv('HOME', '/root')}/.config/gramine/enclave-key.pem",
    )

    pccs_url_default = guess_pccs_url() or "https://pccs.example.com"
    parser.add_argument(
        "--pccs",
        type=str,
        help=f"URL to the PCCS (default: {pccs_url_default})",
        default=pccs_url_default,
    )

    parser.add_argument(
        "--timeout",
        type=int,
        required=False,
        default=24 * 60,
        help="stop the deployment if the application does not "
        "respond after a delay (in min). (Default: 1440 min)",
    )

    parser.add_argument(
        "--output",
        type=Path,
        required=True,
        help="directory to write the args file",
    )

    parser.set_defaults(func=run)


def run(args) -> None:
    """Run the subcommand."""
    client = get_client_docker()

    if app_container_exists(client, args.name):
        raise AppContainerAlreadyRunning(
            f"Docker container `{args.name}` is already running. "
            "Stop and remove it before respawn it!"
        )

    if not is_port_free(args.port):
        raise PortBusy(f"Port {args.port} is already in-used!")

    workspace = args.output.resolve()

    LOG.info("Extracting the package at %s...", workspace)
    package = CodePackage.extract(workspace, args.package)
    code_config = AppConf.load(
        package.config_path, option=AppConfParsingOption.SkipCloud
    )

    image = load_docker_image(client, package.image_tar)

    docker_config = SgxDockerConfig(
        size=args.size,
        host=args.host,
        port=args.port,
        subject=args.subject,
        subject_alternative_name=args.san,
        app_id=uuid4(),
        expiration_date=int((datetime.today() + timedelta(days=args.days)).timestamp()),
        client_certificate=(
            args.client_certificate.read_text() if args.client_certificate else None
        ),
        app_dir=workspace,
        application=code_config.python_application,
        healthcheck=code_config.healthcheck_endpoint,
        signer_key=args.signer_key,
    )

    run_docker_image(
        client,
        args.name,
        image,
        docker_config,
    )

    with Spinner("Waiting for the configuration server to be ready... "):
        wait_for_conf_server(
            ClockTick(
                period=5,
                timeout=60 * args.timeout,
                message="The configuration server is unreachable!",
            ),
            f"https://localhost:{args.port}",
            False,
            get_running_app_container,
            (
                client,
                args.name,
            ),
        )
    LOG.info("The application is now ready to receive the secrets!")

    # Generate evidence and RA-TLS certificate files
    container: Container = get_app_container(client, args.name)

    collect_evidence_and_certificate(container, args.pccs, args.output)


def run_docker_image(
    client: DockerClient,
    app_name: str,
    image: str,
    docker_config: SgxDockerConfig,
):
    """Run the container."""
    LOG.info("Starting the docker...")
    container = client.containers.run(
        image,
        name=app_name,
        command=docker_config.cmd(),
        volumes=docker_config.volumes(),
        devices=SgxDockerConfig.devices(),
        ports=docker_config.ports(),
        entrypoint=SgxDockerConfig.entrypoint,
        labels=docker_config.labels(),
        remove=False,
        detach=True,
        stdout=True,
        stderr=True,
    )

    if container.status != "created":
        raise AppContainerError(
            f"Can't create the container: {container.status} - {container.logs()}"
        )
